#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <assert.h>

#ifdef NDEBUG
# error "no way"
#endif

#define TEMPDIR "/tmp/foo42"

static void setup_testdir()
{
	assert(chdir("/") == 0);
	assert(system("rm -rf "TEMPDIR" && mkdir "TEMPDIR) == 0);
	assert(chdir(TEMPDIR) == 0);
	assert(system("mkdir dir && touch file") == 0);
}

static void afunix_bind(const char *path)
{
	struct sockaddr_un sa;
	int sk, saved_errno;

	sa.sun_family = AF_UNIX;
	strcpy(sa.sun_path, path);
	assert((sk = socket(AF_UNIX, SOCK_STREAM, 0)) >= 0);
	bind(sk, (struct sockaddr *) &sa, sizeof(sa_family_t)+strlen(path)+1);

	saved_errno = errno;
	assert(close(sk) == 0);
	errno = saved_errno;
}

static int careful_rename(const char *old, const char *new)
{
#ifndef DO_IT_ANYWAY
	/* Makes the system crash :-( */
	if(strstr(old, "dir") && strcmp(new, "/") == 0) {
		printf("rename(\"%s\", \"%s\"): don't even think about it.\n", old, new);
		errno = EGRATUITOUS;
		return -1;
	}
#endif
	return rename(old, new);
}

int main(int argc, char **argv)
{
	char *paths[] = {
	       	"",
		"/",
	       	"file",
	       	"file/",
	       	"dir",
	       	"dir/",
		"nosuch",
		"nosuch/",
		NULL
       	};
	char **p, **q;

#define TEST(func, ...) { \
	    setup_testdir(); \
	    errno = 0; \
       	    func(__VA_ARGS__); \
	}
#define TEST1(func, ...) { \
	    TEST(func, __VA_ARGS__); \
	    printf(#func "(\"%s\"): %s\n", *p, strerror(errno)); \
	    fflush(stdout); \
	}
#define TEST2(func, ...) { \
	    TEST(func, __VA_ARGS__); \
	    printf(#func "(\"%s\", \"%s\"): %s\n", *p, *q, strerror(errno)); \
	    fflush(stdout); \
	}

#define TESTS1(func, ...) \
	for(p=paths; *p; p++) TEST1(func, __VA_ARGS__);
#define TESTS2(func, ...) \
	for(p=paths; *p; p++) for(q=paths; *q; q++) TEST2(func, __VA_ARGS__);

	TESTS1(afunix_bind, *p);
	TESTS1(unlink, *p);
	TESTS1(unlinkat, AT_FDCWD, *p, 0);
	TESTS1(mkdir, *p, 0777);
	TESTS1(mkdirat, AT_FDCWD, *p, 0777);
	TESTS1(rmdir, *p);
	TESTS1(symlink, "whatever", *p);
	TESTS1(symlinkat, "whatever", AT_FDCWD, *p);

	TESTS2(link, *p, *q);
	TESTS2(linkat, AT_FDCWD, *p, AT_FDCWD, *q, 0);
	TESTS2(careful_rename, *p, *q);
	/* renameat() is not implemented */

	return 0;
}

